#ifndef art_Framework_Principal_Run_h
#define art_Framework_Principal_Run_h

// ======================================================================
//
// Run - This is the primary interface for accessing per run EDProducts
//       and inserting new derived products.
//
// For its usage, see "art/Framework/Principal/DataViewImpl.h"
// ======================================================================

#include "art/Framework/Principal/DataViewImpl.h"
#include "art/Framework/Principal/fwd.h"
#include "art/Persistency/Provenance/RunAuxiliary.h"
#include "art/Persistency/Provenance/RunID.h"
#include "art/Utilities/TypeID.h"
#include "cpp0x/memory"
#include "cpp0x/utility"

class art::Run : private art::DataViewImpl {
public:
  Run(RunPrincipal& rp, const ModuleDescription& md);
  ~Run(){}

  typedef DataViewImpl Base;
  // AUX functions.
  RunID const& id() const {return aux_.id();}
  RunNumber_t run() const {return aux_.run();}
  Timestamp const& beginTime() const {return aux_.beginTime();}
  Timestamp const& endTime() const {return aux_.endTime();}

  using Base::get;
  using Base::getByLabel;
  using Base::getMany;
  using Base::getManyByType;
  using Base::me;
  using Base::processHistory;

#ifndef __GCCXML__
  ///Put a new product.
  template <typename PROD>
  void
  put(std::unique_ptr<PROD> && product) {put<PROD>(std::move(product), std::string());}

  ///Put a new product with a 'product instance name'
  template <typename PROD>
  void
  put(std::unique_ptr<PROD> && product, std::string const& productInstanceName);
#endif

  // Return true if this Run has been subjected to a process with
  // the given processName, and false otherwise.
  // If true is returned, then ps is filled with the ParameterSets
  // (possibly more than one) used to configure the identified
  // process(es). Equivalent ParameterSets are compressed out of the
  // result.
  bool
  getProcessParameterSet(std::string const& processName,
                         std::vector<fhicl::ParameterSet>& ps) const;

  template <typename PROD>
  bool removeCachedProduct(Handle<PROD> & h) const;

private:
  RunPrincipal const&
  runPrincipal() const;

  RunPrincipal &
  runPrincipal();

  // commit_() is called to complete the transaction represented by
  // this DataViewImpl. The friendships required are gross, but any
  // alternative is not great either.  Putting it into the
  // public interface is asking for trouble
  friend class InputSource;
  friend class DecrepitRelicInputSourceImplementation;
  friend class EDFilter;
  friend class EDProducer;

  void commit_();

  void removeCachedProduct_(BranchID const & bid) const;

  RunAuxiliary const& aux_;
};

#ifndef __GCCXML__
template <typename PROD>
void
art::Run::put(std::unique_ptr<PROD> && product, std::string const& productInstanceName)
{
  if (product.get() == 0) {                // null pointer is illegal
    TypeID typeID(typeid(PROD));
    throw art::Exception(art::errors::NullPointerError)
      << "Run::put: A null unique_ptr was passed to 'put'.\n"
      << "The pointer is of type " << typeID << ".\n"
      << "The specified productInstanceName was '" << productInstanceName << "'.\n";
  }

  BranchDescription const& desc =
    getBranchDescription(TypeID(*product), productInstanceName);

  Wrapper<PROD> *wp(new Wrapper<PROD>(std::move(product)));

  putProducts().push_back(std::make_pair(wp, &desc));

  // product.release(); // The object has been copied into the Wrapper.
  // The old copy must be deleted, so we cannot release ownership.
}

template <typename PROD>
bool
art::Run::removeCachedProduct(Handle<PROD> & h) const {
  bool result { false };
  if (!h.provenance()->produced()) {
    removeCachedProduct_(h.provenance()->branchID());
    h.clear();
    result = true;
  }
  return result;
}

#endif
#endif /* art_Framework_Principal_Run_h */

// Local Variables:
// mode: c++
// End:
